/* field.c */

#include <stdio.h>
#include <search.h>
#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "qelem.h"
#include "tab.h"
#include "points.h"
#include "field.h"
#include "manifold.h"
#include "pull.h"

#include <gtk/gtk.h>
#include "manifold.h"

/* number of queued fields */
int num_queued_fields = 0;

/* waver velosity */
double waver_velosity = WAVER_VELOSITY;

/* strange constant */
double strange_constant = STRANGE_CONSTANT;

/* add new elements at end, and take them out at front when done */
struct qelem field_list = {
  &field_list,
  &field_list };

int news = 0;

/* create and queue up a new field
 *
 * x,y - location of the field
 * mass - its mass
 * particle_number - a unique particle number
 * flag - the flag to use
 *
 */
int new_field ( double x, double y, double mass, int particle_number, int flag, double radius )
{
  FIELD_PTR f;
  static unsigned int field_number = 0;

  f = (FIELD_PTR)malloc(sizeof(FIELD));
  assert(NULL!=f);
  memset(f,0,sizeof(FIELD));

  f->origin_x        = floor(x) + 0.5;
  f->origin_y        = floor(y) + 0.5;
  f->current_radius  = radius;
  f->mass            = mass;
  f->field_number    = ++field_number;
  f->particle_number = particle_number;
  f->del             = 0;
  f->new_flag        = 1;
  f->flag            = flag;

  insque ( f,field_list.q_back );
  num_queued_fields++;

  news++;

  /* return a unique field number */
  return f->field_number;
}

/* remove from queue and destroy a field */
void destroy_field ( FIELD_PTR f )
{
#if 0
  int sts;

  WAVER w;

  /* dump field down into manifold */
  w.x  = f->origin_x;
  w.y  = f->origin_y;
  w.px = f->origin_x;
  w.py = f->origin_y;
  w.v  = 100;
  w.m  = f->mass;
  do {
    sts = waver_propogate( &w );
  } while ( 0 != sts );
#endif

  remque ( f );
  num_queued_fields--;
  free(f);
}

/* tick a field */
int tick_a_field ( FIELD_PTR f, double delta_time )
{
  int i;
  double x,y;
  int cnt = 0;

  if ( POINT_FLAG_UNIFORCE & f->flag ) {
    /* uniforce fields don't grow */
    return 1;
  }

  /* expand field radius by 'waver_velosity' */
  f->current_radius += (delta_time * waver_velosity );

  /* return to caller */
  return 1;

  /* do for each tick in circle */
  for ( i = 0 ; i < CIRCS ; i += 45 ) {

    x = f->origin_x + f->current_radius * rads[i].x;
    y = f->origin_y + f->current_radius * rads[i].y;

    /* out of range if too small ??? */
    if ( x < 0.0 || y < 0.0 ) {
      /* yes */
      continue;
    }
    /* out of range if too big ??? */
    if ( (unsigned int)x >= WIN_SIZE_X || (unsigned int)y >= WIN_SIZE_Y ) {
      /* yes */
      continue;
    }

    /* increment number in range */
    cnt++;

    /* one point on the grid is good enough to keep the field running */
    return cnt;
  }

  /* return number of in range points */
  return cnt;
}

/* tick all fields */
int tick_all_fields ( double delta_time )
{
  int cnt = 0;
  FIELD_PTR f;
  int sts;
  int dels = 0;

  f = (FIELD_PTR)field_list.q_forw;

  while ( (char *)f != (char *)&field_list ) {

    /* is it brand new ??? */
    if ( f->new_flag ) {
      /* yes, skip */
      goto nxt;
    }

    sts = tick_a_field ( f, delta_time );

#if 0
    printf("sts = %d, field_number = %d, r = %f\n",
	   sts,
	   f->field_number,
	   f->current_radius );
#endif

    /* Has this wave propogated outside
     * the test area ???
     */
    if ( !sts && (int)f->mass == -1 ) {
      int tmpx,tmpy;
      FIELD_PTR f1;

      tmpx = f->origin_x;
      tmpy = f->origin_y;
      f1   = (FIELD_PTR)field_list.q_forw;

      /* yes, mark for destruction as it's too far out */
      f->del = 1;
      dels++;

      /* mark first one at these coordinates with +1.0 mass deletable */
      while ( (char *)f1 != (char *)&field_list ) {
	if ( (int)f1->origin_x == tmpx &&
	     (int)f1->origin_y == tmpy &&
	     f1->mass          == 1.0  &&
	     f1->del           == 0 ) {
	  f1->del = 1;
	  dels++;
	  break;
	}
	f1 = (FIELD_PTR)f1->q.q_forw;
      }
    } else {
      /* no, just update count */
      cnt++;
    }

  nxt:;
    /* on to next field */
    f = (FIELD_PTR)f->q.q_forw;
  }

  /* now, delete any if necessary */
  if ( dels ) {
    FIELD_PTR f_tmp;

    f = (FIELD_PTR)field_list.q_forw;
    while ( (char *)f != (char *)&field_list ) {

      /* save next */
      f_tmp = (FIELD_PTR)f->q.q_forw;

      /* should this one be deleted ??? */
      if ( f->del ) {
	/* yes, do so */
	destroy_field ( f );
      }

      /* onward */
      f = f_tmp;
    }
  }

  /* return number of active fields */
  return cnt;
}

/* get this force
 *
 * p     - a point
 * f     - some field
 * dist  - distance between p and f
 * fx,fy - return values here
 *
 */

void get_this_force (POINT_PTR p, FIELD_PTR f, double dist_sq, double *a_fx, double *a_fy)
{
  double h;
  double dx,dy;
  double slope;
  double fx,fy;

  if ( POINT_FLAG_UNIFORCE & f->flag ) {

    h = 0.01; /* force is always a constant */

  } else {

    /* limit force for model */
    if ( dist_sq < 1.0 ) {
      dist_sq = 1.0;
    }

    /* get magnitude of force between 'p' and 'f' */
    h = strange_constant / dist_sq;
  }

  dx = fabs(p->x - f->origin_x);
  dy = fabs(p->y - f->origin_y);

  if ( dx == 0.0 ) {
    fx = 0.0;
    fy = h;
  } else {
    if ( dy == 0.0 ) {
      fy = 0.0;
      fx = h;
    } else {
      slope = dy/dx;
      fx = h / pow( 1.0 + slope*slope, 0.5 );
      fy = slope * fx;
    }
  }

  /* adjust for mass */
  fx *= (p->mass * f->mass);
  fy *= (p->mass * f->mass);

  /* adjust signs */
  if ( f->origin_x < p->x ) {
    fx *= -1.0;
  }
  if ( f->origin_y < p->y ) {
    fy *= -1.0;
  }

  /* do we flip ??? */
  if ( f->flag & 0x2 ) {
    /* yes, we might based on the angle */
    double ff;
    double alpha;

    alpha = atan( fabs(dy/dx) );
    ff = pull_or_push ( alpha );
    fx *= ff;
    fy *= ff;
  }

  *a_fx = fx;
  *a_fy = fy;

#if 0
  printf("m = %f, h = %f, p = [%3.1f,%3.1f], f = [%3.1f,%3.1f] = %f %f\n",
	 f->mass,
	 h,
	 p->x,p->y,
	 f->origin_x,f->origin_y,
	 fx,fy);
#endif
}

#if 0
void get_this_force (POINT_PTR p, FIELD_PTR f, double dist, double *fx, double *fy)
{
  double h;
  double dx,dy;
  double alpha;

  /* limit force for model */
  if ( dist < 1.0 ) {
    dist = 1.0;
#if 0
    *fx = 0.0;
    *fy = 0.0;
    return;
#endif
  }

  /* get magnitude of force between 'p' and 'f' */
  h = strange_constant / ( dist * dist );

  dx = f->origin_x - p->x;
  dy = f->origin_y - p->y;

  alpha = atan( fabs(dy/dx) );
  *fx = p->mass * f->mass * h * cos ( alpha );
  *fy = p->mass * f->mass * h * sin ( alpha );

  /* adjust signs */
  if ( f->origin_x < p->x ) {
    *fx *= -1.0;
  }
  if ( f->origin_y < p->y ) {
    *fy *= -1.0;
  }

  /* do we flip ??? */
  if ( f->flag & 0x2 ) {
    /* yes, we might based on the angle */
    double ff;

    ff = pull_or_push ( alpha );
    *fx *= ff;
    *fy *= ff;
  }

#if 0
  printf("alpha = %f, m = %f, h = %f, p = [%3.2f,%3.2f], f = [%3.2f,%3.2f] = %f %f\n",
	 alpha,f->mass,h,
	 p->x,p->y,
	 f->origin_x,f->origin_y,
	 *fx,*fy);
#endif

}
#endif

/* get forces at point 'p' from all fields */
int get_all_forces ( POINT_PTR p, double *fx, double *fy, int trace_flag )
{
  FIELD_PTR f;
  double dist_sq;
  double tx = 0.0,ty = 0.0;
  double ttx,tty;

  if ( trace_flag ) {
    printf("get_all_forces: entry\n");
  }

  /* get first field in linked list */
  f = (FIELD_PTR)field_list.q_forw;

  while ( (char *)f != (char *)&field_list ) {
    /* process this field */

    /* skip if new */
    if ( f->new_flag ) {
      goto nxt;
    }

#if 1
    /* skip if our own field */
    if ( p->particle_number == f->particle_number ) {
      goto nxt;
    }
#endif

#if 0
    /* skip if in current box */
    if ( (int)x == (int)f->origin_x &&
	(int)y == (int)f->origin_y ) {
      goto nxt;
    }
#endif

    /* is this point within the radius of a force ??? */

#define T1 (p->x - f->origin_x)
#define T2 (p->y - f->origin_y)

    dist_sq = T1*T1 + T2*T2;

#undef T1
#undef T2

    if ( trace_flag ) {
      printf("get_all_forces: dist_sq = %f, current_radius = %f, mass = %d\n",
	     dist_sq,
	     f->current_radius,
	     (int)f->mass);
    }

#if 0
    if ( dist_sq < 1.0 ) {
      goto nxt;
    }
#endif

    if ( (f->current_radius * f->current_radius) < dist_sq ) {
      /* no, skip */
      goto nxt;
    }

    /* update for just these forces */
    ttx = 0.0;
    tty = 0.0;

    get_this_force (p,f,dist_sq,&ttx,&tty);

    /* update total force */
    tx += ttx;
    ty += tty;

  nxt:;
    /* onward */
    f = (FIELD_PTR)f->q.q_forw;
  }

  /* return results to caller */
  *fx = tx;
  *fy = ty;

  if ( trace_flag ) {
    printf("get_all_forces: point [%3d,%3d] forces = %f %f\n",
	   (int)p->x,(int)p->y,
	   tx,ty);
  }

  /* return to caller */
  return 1;
}

/* find a pared field */
FIELD_PTR find_pair ( FIELD_PTR f )
{
  FIELD_PTR f_tmp = (FIELD_PTR)f->q.q_back;

  while ( (char *)f_tmp != (char *)&field_list ) {

    if ( f_tmp->del ) {
      goto nxt;
    }

    if ( f->mass    == -1.0*f_tmp->mass &&
	(int)f->origin_x == (int)f_tmp->origin_x  &&
	(int)f->origin_y == (int)f_tmp->origin_y ) {
      /* found */
      return f_tmp;
    }

  nxt:;
    f_tmp = (FIELD_PTR)f_tmp->q.q_back;
  }

  /* none found */
  return NULL;
}

/* pair up fields, return count of them */
int pair_fields ( void )
{
  FIELD_PTR f;
  int sts = 0;

  f = (FIELD_PTR)field_list.q_forw;

  while ( (char *)f != (char *)&field_list ) {

    f->pair = NULL;

    /* is it already deleted ??? */
    if ( f->del ) {
      /* yes */
      goto nxt;
    }

    /* only consider minus masses */
    if ( f->mass >= 0.0 ) {
      /* skip */
      goto nxt;
    }

    f->pair = find_pair ( f );
    if ( f->pair ) {
      sts++;
    }

  nxt:;
    f = (FIELD_PTR)f->q.q_forw;
  }

  return sts;
}

/* optomize number of fields */
int opt_fields ( void )
{
  FIELD_PTR f;
  FIELD_PTR f_tmp;
  int sts;
  int i;
  int res = 0;

  /* are there any pairs to consider ??? */
  sts = pair_fields ( );

#if 0
  printf("opt_fields: %d pairs found\n",
	 sts);
#endif

  if ( 0 == sts ) {
    /* no, done */
    return res;
  }

  /* walk list and find a pair */
  f = (FIELD_PTR)field_list.q_forw;
  while ( (char *)f != (char *)&field_list ) {

    /* is it paired ??? */
    if ( NULL == f->pair ) {
      /* no, skip */
      goto nxt;
    }

    /* for each point */
    for ( i = 0 ; i < pts_idx ; i++ ) {
      double dist;
      POINT_PTR p;

      /* get pointer to point */
      p = &pts[i];

      /* is this point within the radius of a force ??? */
      dist = sqrt ( (p->x - f->origin_x)*(p->x - f->origin_x) + (p->y - f->origin_y)*(p->y - f->origin_y) );

      if ( f->current_radius <= dist ) {
	/* skip */
	goto nxt;
      }
    }

    /* here, all points are inside force radius, mark pair for later deletion */
    f->del       = 1;
    f->pair->del = 1;
    res++;

  nxt:;
    f = (FIELD_PTR)f->q.q_forw;
  }

  /* process any deletes */
  f = (FIELD_PTR)field_list.q_forw;
  while ( (char *)f != (char *)&field_list ) {

    /* save next */
    f_tmp = (FIELD_PTR)f->q.q_forw;

    /* should this one be deleted ??? */
    if ( f->del ) {
      /* yes, do so */
      destroy_field ( f );
    }

    /* onward */
    f = f_tmp;
  }

  /* return to caller */
  return res;
}

/* get rid of new flag */
void no_new ( void )
{
  FIELD_PTR f;

  if ( news ) {

    news = 0;

    f = (FIELD_PTR)field_list.q_forw;
    while ( (char *)f != (char *)&field_list ) {

      /* no more new flag */
      f->new_flag = 0;

      /* onward */
      f = (FIELD_PTR)f->q.q_forw;
    }
  }
}
